﻿
// ReSharper disable PossiblyUnassignedProperty
// ReSharper disable InconsistentNaming
var Emitter = require("events").EventEmitter;
var Connector = new Emitter();
var fs = require("fs");
var path_module = require("path");

// ReSharper restore InconsistentNaming
// ReSharper restore PossiblyUnassignedProperty

var Conftool;
try {
    var ConftoolPath = "";
    console.log("Environment Variables: " +
        process.env.ARTDAQ_DATABASE_LIB +
        ", " +
        process.env.ARTDAQ_DATABASE_DIR +
        ", " +
        process.env.MRB_BUILDDIR);
    if (!process.env.ARTDAQ_DATABASE_LIB && process.env.ARTDAQ_DATABASE_DIR && process.env.MRB_BUILDDIR) {
        ConftoolPath = path_module.join(process.env["MRB_BUILDDIR"], "artdaq_database", "lib", "conftoolg");
    } else {
        ConftoolPath = path_module.join(process.env["ARTDAQ_DATABASE_LIB"], "node_modules", "conftoolg");
    }
    console.log("Looking for conftoolg module in " + ConftoolPath);
    Conftool = require(ConftoolPath);
    console.log("Module loaded.");
    Conftool.enable_trace();
} catch (E) {
    console.log("Error loading conftoolg module...have you setup the artdaq_database product?");
    console.log(E);
}

/**
 * Determines the appropriate value of the ARTDAQ_DATABASE_URI environment variable used by conftool
 * @param {Object} dbConfig Configuration for database access
 * @param {string} dbConfig.dbprovider Name of the database provider. Currently "filesystem" and "mongo" are supported.
 * @param {string} dbConfig.baseDir Base directory for filesystemdb
 * @param {string} dbConfig.instanceName Name of the database within the provider to access
 * @param {Object} dbConfig.mongoConfig Configuration for mongodb provider
 * @param {string} dbConfig.mongoConfig.dbUser Username for mongodb
 * @param {string} dbConfig.mongoConfig.dbHost Hostname that mongodb is running on
 * @param {int} dbConfig.mongoConfig.port Port that mongodb is running on
 */
function SetupUri(dbConfig) {
    console.log("Setting ARTDAQ_DATABASE_URI using configuration " + JSON.stringify(dbConfig, null, 4));
    if (dbConfig.dbprovider === "filesystem") {
        process.env["ARTDAQ_DATABASE_URI"] = "filesystemdb://" + dbConfig.baseDir + "/filesystemdb/" + dbConfig.instanceName;
    } else if (dbConfig.dbprovider === "mongo") {
        var passStr = "";
        if (fs.existsSync(".dbpasswd")) {
            passStr = ":" + fs.readFileSync(".dbpasswd");
        } else {
            console.log("No Password file (\".dbpasswd\") found. If auth fails, please place this file in the same directory as serverbase.js");
        }
        var userStr = "";
        if (dbConfig.mongoConfig.dbUser !== "") {
            userStr = "@";
        }
        process.env["ARTDAQ_DATABASE_URI"] = "mongodb://" + dbConfig.mongoConfig.dbUser + passStr + userStr + dbConfig.mongoConfig.dbHost + ":" + dbConfig.mongoConfig.dbPort + "/" + dbConfig.instanceName;
    } else {
        console.log("Connector.SetConfig: ERROR: Unrecognized database provider");
    }
    console.log("ARTDAQ_DATABASE_URI is " + process.env["ARTDAQ_DATABASE_URI"]);
}

/**
 * Runs a findconfigs query against the database
 * @returns {Object} JSON object returned by the database
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunGetConfigsQuery = function (dbConfig) {
    console.log("Running findconfigs query");
    SetupUri(dbConfig);
    var query = {
        filter: {},
        operation: "findconfigs",
        dataformat: "gui"
    };
    if (dbConfig.configNameFilter.length === 0) {
        dbConfig.configNameFilter = "*";
    }

//    if (dbConfig.dbprovider === "mongo") {
//        if (dbConfig.configNameFilter === "*") { dbConfig.configNameFilter = ".*"; }

//        query.filter = {
//            "configurations.name": {
//                "$regex": dbConfig.configNameFilter
//           }
//        };
//    } else {
        query.filter = {
            "configurations.name": dbConfig.configNameFilter
        };
//    }

    var output = Conftool.find_global_configurations_ui(JSON.stringify(query, null, 4));
    console.log("findconfigs output: " + JSON.stringify(output, null, 4));
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    // console.log(retjson);
    return JSON.parse(output.second);
};

/**
 * Finds the entities currently present in the database
 * @returns {Object} JSON object returned by the database 
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunGetEntitiesQuery = function (dbConfig) {
    console.log("Running findentities query");
    SetupUri(dbConfig);
    var query = {
        operation: "findentities",
        dataformat: "gui"
    };
    
    var output = Conftool.find_configuration_entities_ui(JSON.stringify(query, null, 4));
    console.log("Findentities output: " + JSON.stringify(JSON.parse(output.second), null, 4));
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    return JSON.parse(output.second);
};

/**
 * Runs the findversions query using the given entity query
 * @param {Object} query - The query to run (as returned by buildfilter)
 * @returns {Object} JSON object returned by the database
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunGetVersionsQuery = function (dbConfig, query) {
    console.log("Running findversions query: " + JSON.stringify(query, null, 4));
    SetupUri(dbConfig);

    var output = Conftool.find_configuration_versions_ui(JSON.stringify(query, null, 4));
    console.log("findversions output: " + JSON.stringify(output, null, 4));
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    return JSON.parse(output.second);
};

/**
 * Runs a buildfilter operation against the database
 * @param {string} configName - Name of the configuration
 * @param {Object} queryIn - Optional query object
 * @returns {Object} JSON object returned by the database
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunBuildFilterQuery = function (dbConfig, configName, queryIn) {
    SetupUri(dbConfig);
    var query = { filter: { "configurations.name": "" } };

    if (queryIn === undefined) {
        console.log("Getting query for " + configName);
        var configs = Connector.RunGetConfigsQuery(dbConfig).search;
        for (var conf in configs) {
            if (configs.hasOwnProperty(conf)) {
                var config = configs[conf];
                if (config.name === configName) {
                    query = config.query;
                }
            }
        }
    } else {
        query = queryIn;
    }

    console.log("Running buildfilter query: " + JSON.stringify(query, null, 4));
    var output = Conftool.build_global_configuration_search_filter_ui(JSON.stringify(query, null, 4));

    for (var file in output.search) {
        if (output.search.hasOwnProperty(file) && query !== null && query !== undefined) {
            if (output.search[file].query.filter["configurations.name"] !== query.filter["configurations.name"]) {
                output.search.splice(file, 1);
            }
        }
    }

    //console.log("buildfilter output: " + JSON.stringify(output,null,4));
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    console.log("BuildfilterQuery return status " + output.first);
    console.log("BuildFilterQuery data: " + JSON.stringify(JSON.parse(output.second), null, 4));
    return JSON.parse(output.second);
};

/**
 * Creates a new configuration containing the given file versions
 * @returns {Object} JSON object returned by the database
 * @param {Object} query - Query to run
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunNewConfigQuery = function (dbConfig, query) {
    console.log("Running newconfig query: " + JSON.stringify(query, null, 4));
    SetupUri(dbConfig);

    var output = Conftool.create_new_global_configuration_ui(JSON.stringify(query, null, 4));
    console.log("newconfig output: " + JSON.stringify(query, null, 4));

    console.log("RunNewConfigQuery return status " + output.first);
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
};

/**
 * Adds a configuration file to a configuration
 * @param {string} configName - Name of the configuration to modify
 * @param {string} version - Version of the new config file
 * @param {string} collectionName - Collection of new config file
 * @param {Object} entity - Entity information of new config file
 * @param {string} entity.name - Name of the configurable entity
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunAddConfigQuery = function (dbConfig, configName, version, collectionName, entity) {
    console.log("RunAddConfigQuery: configName: " + configName + ", version: " + version + ", collectionName: " + collectionName + ", entity: " + JSON.stringify(entity, null, 4));
    SetupUri(dbConfig);
    var query = {
        filter: {
            "entities.name": entity.name,
            "version": version
        },
        collection: collectionName,
        configuration: configName,
        operation: "addconfig",
        dataformat: "gui"
    };

    var output = Conftool.add_configuration_to_global_configuration_ui(JSON.stringify(query, null, 4));
    //console.log("addconfig output: " + JSON.stringify(output,null,4));

    console.log("RunAddConfigQuery return status " + output.first);
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
};

/**
 * Runs a store query against the database
 * @param {string} data - File to store
 * @param {string} collectionName - Name of the collection
 * @param {string} version - Version identifier
 * @param {string} entity - Configurable Entity name
 * @param {string} type - Type of data ("gui", "fhicl", etc.)
 * @param {string} configName - Name of the configuration to store data in (Optional)
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunStoreQuery = function (dbConfig, data, collectionName, version, entity, dataType, configName) {

    console.log("RunStoreConfigQuery: configName: " + configName + ", collectionName: " + collectionName + ", version: " + version + ", entity: " + entity + ", dataType: " + dataType);
    SetupUri(dbConfig);
    var query = {
        version: "" + version,
        collection: "" + collectionName,
        entity: entity,
        operation: "store",
        dataformat: "" + dataType
        , configuration: "" + configName
    };
    if (configName === undefined) {
        query.configuration = "unspecified";
    }

    console.log("RunStoreQuery: Query: " + JSON.stringify(query,null,4) + ", Data: " + data);
    var output = Conftool.store_configuration_ui(JSON.stringify(query, null, 4), data);

    console.log("RunStoreConfigQuery return status " + output.first);
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    console.log("RunStoreConfigQuery return value " + output.second);
};

/**
 * Loads a configuration file from the database
 * @param {Object} query - The query to run (as returned by buildfilter)
 * @param {Boolean} noparse - If set, will not run JSON.parse on output (for FHiCL)
 * @returns {Object} Object from database
 * @throws DBOperationFailedException: More information in ex.message
 */
Connector.RunLoadQuery = function (dbConfig, query, noparse) {
    console.log("Running load query");
    SetupUri(dbConfig);

    var output = Conftool.load_configuration_ui(JSON.stringify(query, null, 4));
    console.log("Conftool Load output: " + output.first);
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }

    if (noparse) return output.second;

    return JSON.parse(output.second);
};

Connector.RunListDatabasesQuery = function (dbconfig, filterobj) {
    console.log("Running ListDatabases query");
    SetupUri(dbconfig);

    if (!filterobj) {
        filterobj = {};
    }

    var query = {
        filter: filterobj,
        dataformat: "gui",
        operation: "listdatabases"
    };
    var output = Conftool.list_databases(JSON.stringify(query, null, 4));
    console.log("Conftool ListDatabases output: " + JSON.stringify(output, null, 4));
    if (!output.first) {
        throw { name: "DBOperationFailedException", message: output.second };
    }
    
    return JSON.parse(output.second);
}

module.exports = Connector;
